@@ -1,14 +1,19 @@
 """Objects for bringing data from SFTP-hosted CSVs into AGOL Feature Services
 """
 
+import datetime
 import json
 import logging
+import re
 import warnings
 from pathlib import Path
+from time import sleep
+from urllib.error import HTTPError
 
 import arcgis
 import numpy as np
 import pandas as pd
+import requests
 from arcgis.features import GeoAccessor, GeoSeriesAccessor
 
 logger = logging.getLogger(__name__)
@@ -18,11 +23,11 @@ class FeatureServiceInlineUpdater:
     """Updates an AGOL Feature Service with data from a pandas DataFrame
     """
 
-    def __init__(self, gis, dataframe, index_columnn):
+    def __init__(self, gis, dataframe, index_column):
         self.gis = gis
         self.new_dataframe = dataframe
-        self.index_column = index_columnn
-        self.new_data_as_dict = self.new_dataframe.set_index(index_columnn).to_dict('index')
+        self.index_column = index_column
+        self.new_data_as_dict = self.new_dataframe.set_index(index_column).to_dict('index')
         self._class_logger = logging.getLogger(__name__).getChild(self.__class__.__name__)
 
     def update_existing_features_in_feature_service_with_arcpy(self, feature_service_url, fields):
@@ -54,7 +59,7 @@ class FeatureServiceInlineUpdater:
                         rows_updated += 1
                     except RuntimeError as error:
                         if 'The value type is incompatible with the field type' in str(error):
-                            raise ValueError('Field type mistmatch between dataframe and feature service') from error
+                            raise ValueError('Field type mismatch between dataframe and feature service') from error
         self._class_logger.info('%s rows updated', rows_updated)
 
         return rows_updated
@@ -179,7 +184,7 @@ class FeatureServiceInlineUpdater:
                 self._class_logger.debug('New data: %s', data['new_values'])
         if update_failures:
             self._class_logger.warning(
-                'The following %s updates failed. As a result, all successfull updates should have been rolled back.',
+                'The following %s updates failed. As a result, all successful updates should have been rolled back.',
                 len(update_failures)
             )
             for _, data in self._get_old_and_new_values(live_dataframe_as_dict, update_failures).items():
@@ -395,7 +400,7 @@ class FeatureServiceAttachmentsUpdater:
                 result = self.feature_layer.attachments.update(target_oid, attachment_id, filepath)
             except Exception:
                 self._class_logger.error(
-                    'AGOL error while overwritting %s (attachment ID %s) on OID %s with %s',
+                    'AGOL error while overwriting %s (attachment ID %s) on OID %s with %s',
                     old_name,
                     attachment_id,
                     target_oid,
@@ -496,14 +501,235 @@ class FeatureServiceAttachmentsUpdater:
 
 
 class FeatureServiceOverwriter:
-    """Overwrites an AGOL Feature Service with data from a pandas DataFrame and a geometry source (Spatially-enabled
-    Data Frame, feature class, etc)
-
-    To be implemented as needed.
+    """Overwrites an AGOL Feature Service with data from a spatially-enabled DataFrame.
     """
 
+    def __init__(self, gis):
+        self.gis = gis
+        self._class_logger = logging.getLogger(__name__).getChild(self.__class__.__name__)
 
-#: TODO: implement for Rick's fleet stuff
+    def _retry(self, worker_method, *args, **kwargs):
+        """Allows you to retry a function/method three times to overcome network jitters
+
+        Retries worker_method three times (for a total of four tries, including the initial attempt), pausing 2^trycount seconds between each retry. Any arguments for worker_method can be passed in as additional parameters to _retry() following worker_method: _retry(foo_method, arg1, arg2, keyword_arg=3)
+
+        Args:
+            worker_method (callable): The name of the method to be retried (minus the calling parens)
+
+        Raises:
+            error: The final error the causes worker_method to fail after 3 retries
+
+        Returns:
+            various: The value(s) returned by worked_method
+        """
+        tries = 1
+        max_tries = 3
+        delay = 2  #: in seconds
+
+        #: this inner function (closure? almost-closure?) allows us to keep track of tries without passing it as an arg
+        def _inner_retry(worker_method, *args, **kwargs):
+            nonlocal tries
+
+            try:
+                return worker_method(*args, **kwargs)
+
+            #: ArcGIS API for Python loves throwing bog-standard Exceptions, so we can't narrow this down further
+            except Exception as error:
+                if tries <= max_tries:  #pylint: disable=no-else-return
+                    wait_time = delay**tries
+                    self._class_logger.debug(
+                        'Exception "%s" thrown on "%s". Retrying after %s seconds...', error, worker_method, wait_time
+                    )
+                    sleep(wait_time)
+                    tries += 1
+                    return _inner_retry(worker_method, *args, **kwargs)
+                else:
+                    raise error
+
+        return _inner_retry(worker_method, *args, **kwargs)
+
+    @staticmethod
+    def _rename_columns_for_agol(columns):
+        """Replace special characters and spaces with '_' to match AGOL field names
+
+        Args:
+            columns (iter): The new columns to be renamed
+
+        Returns:
+            Dict: Mapping {'original name': 'cleaned_name'}
+        """
+
+        rename_dict = {}
+        for column in columns:
+            rename_dict[column] = re.sub(r'[^a-zA-Z0-9_]', '_', column)
+        return rename_dict
+
+    def _check_fields_match(self, featurelayer, new_dataframe):
+        """Make sure new data doesn't have any extra fields, warn if it doesn't contain all live fields
+
+        Args:
+            featurelayer (arcgis.features.FeatureLayer): Live data
+            new_dataframe (pd.DataFrame): New data
+
+        Raises:
+            RuntimeError: If new data contains a field not present in the live data
+        """
+
+        live_fields = {field['name'] for field in featurelayer.properties['fields']}
+        new_fields = set(new_dataframe.columns)
+        #: Remove SHAPE field from set (live "featurelayer.properties['fields']" does not expose the 'SHAPE' field)
+        try:
+            new_fields.remove('SHAPE')
+        except KeyError:
+            pass
+        new_dif = new_fields - live_fields
+        live_dif = live_fields - new_fields
+        if new_dif:
+            raise RuntimeError(
+                f'New dataset contains the following fields that are not present in the live dataset: {new_dif}'
+            )
+        if live_dif:
+            self._class_logger.warning(
+                'New dataset does not contain the following fields that are present in the live dataset: %s', live_dif
+            )
+
+    def _truncate_existing_data(self, featurelayer, layer_index, itemid):
+        """Remove all existing features from live dataset
+
+        Args:
+            featurelayer (arcgis.features.FeatureLayer): Live data
+            layer_index (int): Layer id within the main feature service
+            itemid (str): AGOL item ID for the main feature service
+
+        Raises:
+            RuntimeError: If the truncate fails
+
+        Returns:
+            pd.DataFrame: The feature layer's data prior to truncating
+        """
+
+        old_data = featurelayer.query(as_df=True)
+        truncate_result = self._retry(featurelayer.manager.truncate, asynchronous=True, wait=True)
+        self._class_logger.debug(truncate_result)
+        if truncate_result['status'] != 'Completed':
+            raise RuntimeError(f'Failed to truncate existing data from layer id {layer_index} in itemid {itemid}')
+        return old_data
+
+    def _replace_nan_series_with_empty_strings(self, dataframe):
+        """Fill all completely empty series with empty strings ('')
+
+        As of arcgis 2.0.1. to_featureset() doesn't handle completely empty series properly (it relies on str;
+        https://github.com/Esri/arcgis-python-api/issues/1281), so we convert to empty strings for the time being.
+
+        Args:
+            dataframe (pd.DataFrame): Data to clean/fix
+
+        Returns:
+            pd.DataFrame: The cleaned data
+        """
+
+        for column in dataframe:
+            if dataframe[column].isnull().all():
+                self._class_logger.debug('Column %s is empty; replacing np.nans with empty strings', column)
+                dataframe[column].fillna(value='', inplace=True)
+        return dataframe
+
+    def _append_new_data(self, target_featurelayer, dataframe, feature_service_item_id, layer_index):
+        """Add new data to live dataset
+
+        Args:
+            target_featurelayer (arcgis.features.FeatureLayer): Live dataset; should be empty
+            dataframe (pd.DataFrame): Spatially-enabled dataframe containing new data. Column names should match live data
+            feature_service_item_id (int): Layer ID within the main feature service
+            layer_index (str): AGOL item ID for the main feature service
+
+        Raises:
+            RuntimeError: If append fails; will attempt to rollback appends that worked.
+
+        Returns:
+            dict: Messages returned from append operation
+        """
+
+        geojson = dataframe.spatial.to_featureset().to_geojson
+        result, messages = self._retry(
+            target_featurelayer.append,
+            upload_format='geojson',
+            edits=geojson,
+            upsert=False,
+            return_messages=True,
+            rollback=True,
+        )
+
+        self._class_logger.debug(messages)
+        if not result:
+            raise RuntimeError(
+                f'Failed to append data to layer id {layer_index} in itemid {feature_service_item_id}. Append should have been rolled back.'
+            )
+
+        return messages
+
+    def _save_truncated_data(self, dataframe, directory):
+        """Save the pre-truncate dataframe to directory for safety
+
+        Args:
+            dataframe (pd.DataFrame): The data extracted from the feature layer prior to truncating
+            directory (str or Path): The directory to save the data to.
+
+        Returns:
+            Path: The full path to the output file, named with today's date.
+        """
+
+        out_path = Path(directory, f'old_data_{datetime.date.today()}.json')
+        out_path.write_text(dataframe.spatial.to_featureset().to_json)
+        return out_path
+
+    def truncate_and_load_feature_service(self, feature_service_item_id, new_dataframe, failsafe_dir, layer_index=0):
+        """Attempt to delete existing data from a feature layer and add new data from a spatially-enabled dataframe.
+
+        First attempts to truncate existing data. Then renames new data column names to conform to AGOL scheme (spaces, special chars changed to '_'). Finally attempts to append new data to now-empty feature layer.
+
+        Args:
+            feature_service_item_id (str): AGOL item ID for feature service to truncate and load
+            new_dataframe (pd.DataFrame): Spatially-enabled dataframe containing new data. Must not contain columns that do not exist in the live data. Geometry type must match existing data.
+            layer_index (int, optional): ID for the feature service layer to be truncated and loaded. Defaults to 0.
+
+        Raises:
+            RuntimeError: If new data contains a field not present in the live data
+            RuntimeError: If the truncate fails
+            RuntimeError: If append fails; will attempt to rollback appends that worked.
+
+        Returns:
+            int: The number of new records added after deleting all existing records.
+        """
+
+        target_featurelayer = arcgis.features.FeatureLayer.fromitem(
+            self.gis.content.get(feature_service_item_id), layer_id=layer_index
+        )
+        self._class_logger.info('Truncating existing data...')
+        old_dataframe = self._truncate_existing_data(target_featurelayer, layer_index, feature_service_item_id)
+        cleaned_dataframe = new_dataframe.rename(columns=self._rename_columns_for_agol(new_dataframe.columns))
+        #: temp fix until Esri fixes empty series as NaN bug
+        fixed_dataframe = self._replace_nan_series_with_empty_strings(cleaned_dataframe)
+        self._check_fields_match(target_featurelayer, fixed_dataframe)
+        self._class_logger.info('Loading new data...')
+        try:
+            messages = self._append_new_data(target_featurelayer, fixed_dataframe, feature_service_item_id, layer_index)
+        except Exception:
+            try:
+                self._class_logger.info('Append failed; attempting to re-load truncated data...')
+                messages = self._append_new_data(
+                    target_featurelayer, old_dataframe, feature_service_item_id, layer_index
+                )
+                self._class_logger.info('%s features reloaded', messages['recordCount'])
+            except Exception as inner_error:
+                failsafe_path = self._save_truncated_data(old_dataframe, failsafe_dir)
+                raise RuntimeError(
+                    f'Failed to re-add truncated data after failed append; data saved to {failsafe_path}'
+                ) from inner_error
+            finally:
+                raise
+
+        return messages['recordCount']
 
 
 class ColorRampReclassifier:
