@@ -1,6 +1,8 @@
 import builtins
 import json
 import logging
+import sys
+import urllib
 from pathlib import Path
 
 import arcgis
@@ -8,6 +10,7 @@ import numpy as np
 import pandas as pd
 import pandas.testing as tm
 import pytest
+import requests
 from arcgis.features import GeoAccessor, GeoSeriesAccessor
 from mock_arcpy import arcpy
 from pandas.api.types import CategoricalDtype
@@ -407,7 +410,7 @@ class TestFeatureServiceInlineUpdaterResultParsing:
 
             assert caplog.records[
                 0
-            ].message == 'The following 1 updates failed. As a result, all successfull updates should have been rolled back.'
+            ].message == 'The following 1 updates failed. As a result, all successful updates should have been rolled back.'
             assert caplog.records[0].levelname == 'WARNING'
 
             assert caplog.records[1].message == "Existing data: {'objectId': 2, 'data': 'bar'}"
@@ -526,7 +529,7 @@ class TestFeatureServiceInlineUpdaterResultParsing:
             assert caplog.records[2].levelname == 'DEBUG'
             assert caplog.records[
                 3
-            ].message == 'The following 1 updates failed. As a result, all successfull updates should have been rolled back.'
+            ].message == 'The following 1 updates failed. As a result, all successful updates should have been rolled back.'
             assert caplog.records[3].levelname == 'WARNING'
             assert caplog.records[4].message == "Existing data: {'objectId': 2, 'data': 'bar'}"
             assert caplog.records[4].levelname == 'WARNING'
@@ -1567,7 +1570,7 @@ class TestAttachments:
 
         count = updater._overwrite_attachments_by_oid(action_df, 'path')
         assert count == 1
-        assert 'AGOL error while overwritting oldname1 (attachment ID 11) on OID 1 with path1' in caplog.text
+        assert 'AGOL error while overwriting oldname1 (attachment ID 11) on OID 1 with path1' in caplog.text
         assert 'foo' in caplog.text
         assert updater.failed_dict == {1: ('update', 'path1')}
 
@@ -1646,3 +1649,459 @@ class TestAttachments:
         tm.assert_frame_equal(attachment_df[other_fields], test_df[other_fields])
         assert [str(path) for path in attachment_df['full_file_path']
                ] == [str(path) for path in test_df['full_file_path']]
+
+
+class TestFeatureServiceOverwriter:
+
+    def test_rename_columns_from_agol_handles_special_and_space(self):
+        cols = ['Test Name:']
+
+        renamed = palletjack.FeatureServiceOverwriter._rename_columns_for_agol(cols)
+
+        assert renamed == {'Test Name:': 'Test_Name_'}
+
+    def test_rename_columns_from_agol_handles_special_and_space_leaves_others_alone(self):
+        cols = ['Test Name:', 'FooName']
+
+        renamed = palletjack.FeatureServiceOverwriter._rename_columns_for_agol(cols)
+
+        assert renamed == {'Test Name:': 'Test_Name_', 'FooName': 'FooName'}
+
+    def test_rename_columns_from_agol_retains_underscores(self):
+        cols = ['Test Name:', 'Foo_Name']
+
+        renamed = palletjack.FeatureServiceOverwriter._rename_columns_for_agol(cols)
+
+        assert renamed == {'Test Name:': 'Test_Name_', 'Foo_Name': 'Foo_Name'}
+
+    def test_check_fields_match_normal(self, mocker):
+        mock_fl = mocker.Mock()
+        mock_fl.properties = {
+            'fields': [
+                {
+                    'name': 'Foo'
+                },
+                {
+                    'name': 'Bar'
+                },
+            ]
+        }
+        df = pd.DataFrame(columns=['Foo', 'Bar'])
+        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
+
+        palletjack.FeatureServiceOverwriter._check_fields_match(overwriter, mock_fl, df)
+
+    def test_check_fields_match_raises_error_on_extra_new_field(self, mocker):
+        mock_fl = mocker.Mock()
+        mock_fl.properties = {
+            'fields': [
+                {
+                    'name': 'Foo'
+                },
+                {
+                    'name': 'Bar'
+                },
+            ]
+        }
+        df = pd.DataFrame(columns=['Foo', 'Bar', 'Baz'])
+        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
+
+        with pytest.raises(RuntimeError) as exc_info:
+            palletjack.FeatureServiceOverwriter._check_fields_match(overwriter, mock_fl, df)
+
+        assert exc_info.value.args[
+            0] == 'New dataset contains the following fields that are not present in the live dataset: {\'Baz\'}'
+
+    def test_check_fields_match_ignores_new_shape_field(self, mocker):
+        mock_fl = mocker.Mock()
+        mock_fl.properties = {
+            'fields': [
+                {
+                    'name': 'Foo'
+                },
+                {
+                    'name': 'Bar'
+                },
+            ]
+        }
+        df = pd.DataFrame(columns=['Foo', 'Bar', 'SHAPE'])
+        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
+
+        palletjack.FeatureServiceOverwriter._check_fields_match(overwriter, mock_fl, df)
+
+    def test_check_fields_match_warns_on_missing_new_field(self, mocker, caplog):
+        mock_fl = mocker.Mock()
+        mock_fl.properties = {
+            'fields': [
+                {
+                    'name': 'Foo'
+                },
+                {
+                    'name': 'Bar'
+                },
+                {
+                    'name': 'Baz'
+                },
+            ]
+        }
+        df = pd.DataFrame(columns=['Foo', 'Bar'])
+        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
+
+        palletjack.FeatureServiceOverwriter._check_fields_match(overwriter, mock_fl, df)
+
+        assert 'New dataset does not contain the following fields that are present in the live dataset: {\'Baz\'}' in caplog.text
+
+    def test_truncate_existing_data_normal(self, mocker):
+        fl_mock = mocker.Mock()
+        fl_mock.manager.truncate.return_value = {
+            'submissionTime': 123,
+            'lastUpdatedTime': 124,
+            'status': 'Completed',
+        }
+        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
+
+        palletjack.FeatureServiceOverwriter._truncate_existing_data(overwriter, fl_mock, 0, 'abc')
+
+    def test_truncate_existing_raises_error_on_failure(self, mocker):
+        fl_mock = mocker.Mock()
+        fl_mock.manager.truncate.return_value = {
+            'submissionTime': 123,
+            'lastUpdatedTime': 124,
+            'status': 'Foo',
+        }
+        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
+
+        with pytest.raises(RuntimeError) as exc_info:
+            palletjack.FeatureServiceOverwriter._truncate_existing_data(overwriter, fl_mock, 0, 'abc')
+
+        assert exc_info.value.args[0] == 'Failed to truncate existing data from layer id 0 in itemid abc'
+
+    def test_truncate_existing_retries_on_HTTPError(self, mocker):
+        fl_mock = mocker.Mock()
+        fl_mock.manager.truncate.side_effect = [
+            urllib.error.HTTPError('url', 'code', 'msg', 'hdrs', 'fp'), {
+                'submissionTime': 123,
+                'lastUpdatedTime': 124,
+                'status': 'Completed',
+            }
+        ]
+        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
+
+        palletjack.FeatureServiceOverwriter._truncate_existing_data(overwriter, fl_mock, 0, 'abc')
+
+    def test_retry_returns_on_first_success(self, mocker):
+        mock = mocker.Mock()
+        mock.function.return_value = 42
+
+        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
+
+        answer = overwriter._retry(mock.function, 'a', 'b')
+
+        assert answer == 42
+        assert mock.function.call_count == 1
+
+    def test_retry_returns_after_one_failure(self, mocker):
+        mock = mocker.Mock()
+        mock.function.side_effect = [urllib.error.HTTPError('a', 'b', 'c', 'd', 'e'), 42]
+        mocker.patch('palletjack.updaters.sleep')
+
+        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
+
+        answer = overwriter._retry(mock.function, 'a', 'b')
+
+        assert answer == 42
+        assert mock.function.call_count == 2
+
+    def test_retry_returns_after_two_failures(self, mocker):
+        mock = mocker.Mock()
+        mock.function.side_effect = [
+            urllib.error.HTTPError('a', 'b', 'c', 'd', 'e'),
+            urllib.error.HTTPError('a', 'b', 'c', 'd', 'e'),
+            42,
+        ]
+        mocker.patch('palletjack.updaters.sleep')
+
+        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
+
+        answer = overwriter._retry(mock.function, 'a', 'b')
+
+        assert answer == 42
+        assert mock.function.call_count == 3
+
+    def test_retry_fails_after_four_failures(self, mocker):
+        mock = mocker.Mock()
+        mock.function.side_effect = [
+            urllib.error.HTTPError('a', 'b', 'c', 'd', 'e'),
+            urllib.error.HTTPError('a', 'b', 'c', 'd', 'e'),
+            urllib.error.HTTPError('a', 'b', 'c', 'd', 'e'),
+            urllib.error.HTTPError('a', 'b', 'c', 'd', 'e'),
+            42,
+        ]
+        mocker.patch('palletjack.updaters.sleep')
+
+        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
+
+        with pytest.raises(urllib.error.HTTPError):
+            answer = overwriter._retry(mock.function, 'a', 'b')
+
+        # assert answer == 42
+        assert mock.function.call_count == 4
+
+    def test_retry_retries_on_request_http_error(self, mocker):
+        mock = mocker.Mock()
+        mock.function.side_effect = [requests.exceptions.HTTPError, 42]
+        mocker.patch('palletjack.updaters.sleep')
+
+        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
+
+        answer = overwriter._retry(mock.function, 'a', 'b')
+
+        assert answer == 42
+
+    def test_retry_retries_on_request_ssl_error(self, mocker):
+        mock = mocker.Mock()
+        mock.function.side_effect = [requests.exceptions.SSLError, 42]
+        mocker.patch('palletjack.updaters.sleep')
+
+        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
+
+        answer = overwriter._retry(mock.function, 'a', 'b')
+
+        assert answer == 42
+
+    def test_append_new_data_doesnt_raise_on_normal(self, mocker):
+        mock_df = mocker.Mock()
+        mock_fl = mocker.Mock()
+        mock_fl.append.return_value = (True, {'message': 'foo'})
+
+        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
+        overwriter._append_new_data(mock_fl, mock_df, 0, 'abc')
+
+    def test_append_new_data_retries_on_httperror(self, mocker):
+        mock_df = mocker.Mock()
+        mock_fl = mocker.Mock()
+        mock_fl.append.side_effect = [urllib.error.HTTPError('a', 'b', 'c', 'd', 'e'), (True, {'message': 'foo'})]
+
+        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
+        overwriter._append_new_data(mock_fl, mock_df, 0, 'abc')
+
+    def test_append_new_data_raises_on_False_result(self, mocker):
+        mock_df = mocker.Mock()
+        mock_fl = mocker.Mock()
+        mock_fl.append.return_value = (False, {'message': 'foo'})
+
+        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
+
+        with pytest.raises(RuntimeError) as exc_info:
+            overwriter._append_new_data(mock_fl, mock_df, 'abc', 0)
+
+        assert exc_info.value.args[
+            0] == 'Failed to append data to layer id 0 in itemid abc. Append should have been rolled back.'
+
+    def test_truncate_and_load_feature_service_normal(self, mocker):
+        mock_fl = mocker.Mock()
+        mock_fl.manager.truncate.return_value = {
+            'submissionTime': 123,
+            'lastUpdatedTime': 124,
+            'status': 'Completed',
+        }
+        mock_fl.properties = {
+            'fields': [
+                {
+                    'name': 'Foo'
+                },
+                {
+                    'name': 'Bar'
+                },
+            ]
+        }
+        mock_fl.append.return_value = (True, {'recordCount': 42})
+
+        fl_class_mock = mocker.Mock()
+        fl_class_mock.fromitem.return_value = mock_fl
+        mocker.patch('arcgis.features.FeatureLayer', fl_class_mock)
+
+        new_dataframe = pd.DataFrame(columns=['Foo', 'Bar'])
+        mocker.patch.object(pd.DataFrame, 'spatial')
+
+        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
+
+        uploaded_features = overwriter.truncate_and_load_feature_service('abc', new_dataframe, 'foo/dir')
+
+        assert uploaded_features == 42
+
+    def test_truncate_and_load_append_fails_reload_works(self, mocker, caplog):
+        caplog.set_level(logging.DEBUG)
+        mock_fl = mocker.Mock()
+        mock_fl.manager.truncate.return_value = {
+            'submissionTime': 123,
+            'lastUpdatedTime': 124,
+            'status': 'Completed',
+        }
+        mock_fl.properties = {
+            'fields': [
+                {
+                    'name': 'Foo'
+                },
+                {
+                    'name': 'Bar'
+                },
+            ]
+        }
+        mock_fl.append.side_effect = [(False, 'foo'), (True, {'recordCount': 42})]
+
+        fl_class_mock = mocker.Mock()
+        fl_class_mock.fromitem.return_value = mock_fl
+        mocker.patch('arcgis.features.FeatureLayer', fl_class_mock)
+
+        new_dataframe = pd.DataFrame(columns=['Foo', 'Bar'])
+        mocker.patch.object(pd.DataFrame, 'spatial')
+
+        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
+
+        with pytest.raises(
+            RuntimeError,
+            match='Failed to append data to layer id 0 in itemid abc. Append should have been rolled back.'
+        ):
+            uploaded_features = overwriter.truncate_and_load_feature_service('abc', new_dataframe, 'foo/dir')
+
+        assert 'Append failed; attempting to re-load truncated data...' in caplog.text
+        assert '42 features reloaded' in caplog.text
+
+    def test_truncate_and_load_append_fails_reload_fails(self, mocker, caplog):
+        caplog.set_level(logging.DEBUG)
+        mock_fl = mocker.Mock()
+        mock_fl.manager.truncate.return_value = {
+            'submissionTime': 123,
+            'lastUpdatedTime': 124,
+            'status': 'Completed',
+        }
+        mock_fl.properties = {
+            'fields': [
+                {
+                    'name': 'Foo'
+                },
+                {
+                    'name': 'Bar'
+                },
+            ]
+        }
+        mock_fl.append.side_effect = [(False, 'foo'), (False, 'bar')]
+
+        fl_class_mock = mocker.Mock()
+        fl_class_mock.fromitem.return_value = mock_fl
+        mocker.patch('arcgis.features.FeatureLayer', fl_class_mock)
+
+        new_dataframe = pd.DataFrame(columns=['Foo', 'Bar'])
+        mocker.patch.object(pd.DataFrame, 'spatial')
+
+        mocker.patch.object(palletjack.FeatureServiceOverwriter, '_save_truncated_data', return_value='/foo/bar.json')
+
+        overwriter = palletjack.FeatureServiceOverwriter(mocker.Mock())
+
+        with pytest.raises(RuntimeError) as exc_info:
+            uploaded_features = overwriter.truncate_and_load_feature_service('abc', new_dataframe, 'foo/dir')
+
+        assert exc_info.value.args[
+            0] == 'Failed to re-add truncated data after failed append; data saved to /foo/bar.json'
+        assert 'Append failed; attempting to re-load truncated data...' in caplog.text
+        assert 'features reloaded' not in caplog.text
+
+    def test_replace_nan_series_with_empty_strings_one_empty_one_non_empty_float(self, mocker):
+        df = pd.DataFrame({
+            'normal': [1., 2., 3.],
+            'empty': [np.nan, np.nan, np.nan],
+        })
+
+        fixed_df = palletjack.FeatureServiceOverwriter._replace_nan_series_with_empty_strings(mocker.Mock(), df)
+
+        test_df = pd.DataFrame({
+            'normal': [1., 2., 3.],
+            'empty': ['', '', ''],
+        })
+
+        tm.assert_frame_equal(fixed_df, test_df)
+
+    def test_replace_nan_series_with_empty_strings_other_series_has_nan(self, mocker):
+        df = pd.DataFrame({
+            'normal': [1., 2., np.nan],
+            'empty': [np.nan, np.nan, np.nan],
+        })
+
+        fixed_df = palletjack.FeatureServiceOverwriter._replace_nan_series_with_empty_strings(mocker.Mock(), df)
+
+        test_df = pd.DataFrame({
+            'normal': [1., 2., np.nan],
+            'empty': ['', '', ''],
+        })
+
+        tm.assert_frame_equal(fixed_df, test_df)
+
+    def test_replace_nan_series_with_empty_strings_other_series_is_empty_str(self, mocker):
+        df = pd.DataFrame({
+            'normal': ['', '', ''],
+            'empty': [np.nan, np.nan, np.nan],
+        })
+
+        fixed_df = palletjack.FeatureServiceOverwriter._replace_nan_series_with_empty_strings(mocker.Mock(), df)
+
+        test_df = pd.DataFrame({
+            'normal': ['', '', ''],
+            'empty': ['', '', ''],
+        })
+
+        tm.assert_frame_equal(fixed_df, test_df)
+
+    def test_save_truncated_data_calls_write_with_json(self, mocker):
+        mock_df = pd.DataFrame({
+            'foo': [1],
+            'x': [11],
+            'y': [14],
+        })
+        #: Need to 'unload' the mocked arcpy used in tests above so that .from_xy() works. This could be pulled
+        #: out if these tests or the other tests are in their own files.
+        if 'arcpy' in sys.modules:
+            del sys.modules['arcpy']
+
+        mock_sdf = pd.DataFrame.spatial.from_xy(mock_df, 'x', 'y')
+
+        open_mock = mocker.MagicMock()
+        context_manager_mock = mocker.MagicMock()
+        context_manager_mock.return_value.__enter__.return_value = open_mock
+        mocker.patch('pathlib.Path.open', new=context_manager_mock)
+
+        palletjack.FeatureServiceOverwriter._save_truncated_data(mocker.Mock(), mock_sdf, 'foo')
+
+        test_json_string = '{"features": [{"geometry": {"spatialReference": {"wkid": 4326}, "x": 11, "y": 14}, "attributes": {"foo": 1, "x": 11, "y": 14, "OBJECTID": 1}}], "objectIdFieldName": "OBJECTID", "displayFieldName": "OBJECTID", "spatialReference": {"wkid": 4326}, "geometryType": "esriGeometryPoint", "fields": [{"name": "OBJECTID", "type": "esriFieldTypeOID", "alias": "OBJECTID"}, {"name": "foo", "type": "esriFieldTypeInteger", "alias": "foo"}, {"name": "x", "type": "esriFieldTypeInteger", "alias": "x"}, {"name": "y", "type": "esriFieldTypeInteger", "alias": "y"}]}'
+
+        assert open_mock.write.called_with(test_json_string)
+
+    def test_save_truncated_data_opens_file_with_right_name(self, mocker):
+        mock_df = pd.DataFrame({
+            'foo': [1],
+            'x': [11],
+            'y': [14],
+        })
+        #: Need to 'unload' the mocked arcpy used in tests above so that .from_xy() works. This could be pulled
+        #: out if these tests or the other tests are in their own files.
+        if 'arcpy' in sys.modules:
+            del sys.modules['arcpy']
+
+        mock_sdf = pd.DataFrame.spatial.from_xy(mock_df, 'x', 'y')
+
+        open_mock = mocker.MagicMock()
+        context_manager_mock = mocker.MagicMock()
+        context_manager_mock.return_value.__enter__.return_value = open_mock
+
+        datetime_mock = mocker.Mock()
+        datetime_mock.date.today.return_value = 'foo-date'
+        mocker.patch('palletjack.updaters.datetime', new=datetime_mock)
+
+        open_mock = mocker.MagicMock()
+        context_manager_mock = mocker.MagicMock()
+        context_manager_mock.return_value.__enter__.return_value = open_mock
+        mocker.patch('pathlib.Path.open', new=context_manager_mock)
+
+        out_path = palletjack.FeatureServiceOverwriter._save_truncated_data(mocker.Mock(), mock_sdf, 'foo')
+
+        assert out_path == Path('foo/old_data_foo-date.json')
